from rest_framework.pagination import PageNumberPagination
from collections import OrderedDict
from rest_framework.response import Response


class Pagination(PageNumberPagination):
    page_size_query_param = "limit"

    def paginate_queryset(self, queryset, request, view=None):
        data = super(Pagination, self).paginate_queryset(
            queryset, request, view
        )
        if not view or not getattr(view, "router", None):
            return data

        if request.accepted_renderer.format != "json":
            return data

        non_format_kwargs = [
            kwarg
            for kwarg in list(view.kwargs.keys()) + list(request.GET.keys())
            if kwarg != "format"
        ]
        if view.action != "list" or any(non_format_kwargs):
            return data

        conf = view.router.get_model_config(queryset.model)
        cache = conf.get("cache", "first_page")

        if cache == "first_page":
            return data
        elif cache == "all":
            return list(queryset)
        elif cache == "none":
            return []
        elif cache in ("filter", "autoupdate"):
            cache_filter = view.router.get_cache_filter_for_model(
                queryset.model
            )
            return list(cache_filter(queryset, self.request))

        return data

    def get_paginated_response(self, data):
        return Response(
            OrderedDict(
                [
                    # DRF default metadata
                    ("count", self.page.paginator.count),
                    ("next", self.get_next_link()),
                    ("previous", self.get_previous_link()),
                    # wq.db additional metadata
                    ("page", self.page.number),
                    ("pages", self.page.paginator.num_pages),
                    ("per_page", self.page.paginator.per_page),
                    ("multiple", self.page.paginator.num_pages > 1),
                    # Actual data ('results' in DRF)
                    ("list", data),
                ]
            )
        )



from .base import APITestCase
from rest_framework import status
from tests.rest_app.models import ItemType, DateModel, ChoiceModel
from tests.rest_app.serializers import (
    ChoiceLabelSerializer,
    DateLabelSerializer,
    ItemLabelSerializer,
)


class LabelTestCase(APITestCase):
    def setUp(self):
        itype = ItemType.objects.create(name="Test", pk=1)
        itype.item_set.create(name="Test 1")
        DateModel.objects.create(
            pk=1,
            name="Test",
            date="2015-01-01 12:00Z",
        )
        ChoiceModel.objects.create(
            pk=1,
            name="Test",
            choice="two",
        )

    def test_rest_date_label(self):
        response = self.client.get("/datemodels/1.json")
        self.assertTrue(status.is_success(response.status_code), response.data)
        self.assertIn("date_label", response.data)
        self.assertEqual(response.data["date_label"], "2015-01-01 06:00 AM")

    def test_rest_choice_label(self):
        response = self.client.get("/choicemodels/1.json")
        self.assertTrue(status.is_success(response.status_code), response.data)
        self.assertIn("choice_label", response.data)
        self.assertEqual(response.data["choice_label"], "Choice Two")

    def test_rest_boolean_label(self):
        response = self.client.get("/itemtypes/1.json")
        self.assertTrue(status.is_success(response.status_code), response.data)
        self.assertIn("active_label", response.data)
        self.assertEqual(response.data["active_label"], "Yes")

    def test_custom_date_label(self):
        from wq.db.rest import router

        # Default label
        item = DateModel.objects.get(pk=1)
        obj = router.serialize(item)
        self.assertEqual(obj["date_label"], "2015-01-01 06:00 AM")

        # Custom label
        obj = DateLabelSerializer(item).data
        self.assertEqual(obj["date_label"], "January 1, 2015")

    def test_custom_choice_label(self):
        from wq.db.rest import router

        # Default label
        item = ChoiceModel.objects.get(pk=1)
        obj = router.serialize(item)
        self.assertEqual(obj["choice_label"], "Choice Two")

        # Custom label
        obj = ChoiceLabelSerializer(item).data
        self.assertEqual(obj["choice_label"], "TWO")

    def test_custom_fk_label(self):
        from wq.db.rest import router

        # Default labels
        item = ItemType.objects.get(pk=1).item_set.get(name="Test 1")
        obj = router.serialize(item)
        self.assertEqual(obj["type_id"], 1)
        self.assertEqual(obj["label"], "Test 1")

        # Custom labels
        obj = ItemLabelSerializer(item).data
        self.assertEqual(obj["type_id"], "id-1")
        self.assertEqual(obj["type_label"], "TEST")
        self.assertEqual(obj["label"], "Test: Test 1")



from wq.db import rest
from .models import (
    Campaign,
    Attribute,
    Entity,
    Value,
)

rest.router.register(
    Campaign,
    lookup="slug",
    nested_arrays=Attribute,
    fields="__all__",
)


@rest.register(Attribute, fields="__all__")
class AttributeSerializer(rest.ModelSerializer):
    class Meta:
        wq_field_config = {
            "is_active": {
                "control": {"appearance": "checkbox"},
            },
        }


class ValueSerializer(rest.ModelSerializer):
    class Meta:
        model = Value
        exclude = ("entity",)
        wq_field_config = {
            "attribute": {
                "control": {"appearance": "attribute-label"},
            }
        }


@rest.register(Entity, fields="__all__")
class EntitySerializer(rest.ModelSerializer):
    values = ValueSerializer(
        many=True,
        wq_config={
            "initial": {
                "type_field": "attribute",
                "filter": {
                    "is_active": True,
                    "campaign": "{{campaign_id}}",
                },
            },
        },
    )

    class Meta:
        wq_field_config = {
            "campaign": {
                "control": {"appearance": "campaign-label"},
            }
        }
        wq_fieldsets = {
            "general": {
                "label": "Info",
                "control": {"appearance": "fieldset"},
                "fields": ["name"],
            },
            "values": {
                "label": "Values",
                "control": {"appearance": "eav-fieldset-array"},
            },
        }



import unittest
from .base import APITestCase
from rest_framework import status
import json
from tests.rest_app.models import SlugModel, FieldsetModel
from tests.gis_app.models import GeometryModel
from django.contrib.auth.models import User
from django.conf import settings


class RestPostTestCase(APITestCase):
    def setUp(self):
        self.user = User.objects.create(username="testuser", is_superuser=True)
        self.client.force_authenticate(self.user)

    @unittest.skipUnless(settings.WITH_GIS, "requires GIS")
    def test_rest_geometry_post_geojson(self):
        """
        Posting GeoJSON to a model with a geometry field should work.
        """
        form = {
            "name": "Geometry Test 1",
            "geometry": json.dumps(
                {"type": "Point", "coordinates": [-90, 44]}
            ),
        }

        # Test for expected response
        response = self.client.post("/geometrymodels.json", form)
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )

        # Double-check ORM model & geometry attribute
        obj = GeometryModel.objects.get(id=response.data["id"])
        geom = obj.geometry
        self.assertEqual(geom.srid, 4326)
        self.assertEqual(geom.x, -90)
        self.assertEqual(geom.y, 44)

    @unittest.skipUnless(settings.WITH_GIS, "requires GIS")
    def test_rest_geometry_post_wkt(self):
        """
        Posting WKT to a model with a geometry field should work.
        """
        form = {
            "name": "Geometry Test 2",
            "geometry": "POINT(%s %s)" % (-97, 50),
        }

        # Test for expected response
        response = self.client.post("/geometrymodels.json", form)
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )

        # Double-check ORM model & geometry attribute
        obj = GeometryModel.objects.get(id=response.data["id"])
        geom = obj.geometry
        self.assertEqual(geom.srid, 4326)
        self.assertEqual(geom.x, -97)
        self.assertEqual(geom.y, 50)

    def test_rest_date_label_post(self):
        """
        Posting to a model with a date should return a label and an ISO date
        """
        form = {
            "name": "Test Date",
            "date": "2015-06-01 12:00:00Z",
        }
        response = self.client.post("/datemodels.json", form)
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )
        self.assertIn("date_label", response.data)
        self.assertEqual(response.data["date_label"], "2015-06-01 07:00 AM")
        self.assertIn("date", response.data)
        self.assertEqual(response.data["date"], "2015-06-01T07:00:00-05:00")

    def test_rest_empty_date_post(self):
        """
        Allow posting an empty date if the field allows nulls
        """
        form = {
            "name": "Test Date",
            "date": "2015-06-01 12:00:00Z",
            "empty_date": "",
        }
        response = self.client.post("/datemodels.json", form)
        self.assertEqual(
            response.status_code, status.HTTP_201_CREATED, response.data
        )

    def test_rest_empty_string(self):
        def check_result(form, expected_output):
            response = self.client.post("/charfieldmodels.json", form)
            if not expected_output:
                self.assertEqual(
                    response.status_code,
                    status.HTTP_400_BAD_REQUEST,
                    response.data,
                )
                return

            self.assertEqual(
                response.status_code, status.HTTP_201_CREATED, response.data
            )
            for field, value in expected_output.items():
                if value is None:
                    self.assertIsNone(response.data[field])
                else:
                    self.assertEqual(response.data[field], value)

        # No fields submitted
        check_result({}, False)

        # Required field submitted but left blank
        check_result({"required_field": ""}, False)

        # Required field submitted
        check_result(
            {
                "required_field": "test",
            },
            {
                "required_field": "test",
                "nullable_field": None,
                "blankable_field": "",
                "nullableblankable_field": None,
            },
        )

        # All fields submitted but left blank
        check_result(
            {
                "required_field": "test",
                "nullable_field": "",
                "blankable_field": "",
                "nullableblankable_field": "",
            },
            {
                "required_field": "test",
                "nullable_field": None,
                "blankable_field": "",
                "nullableblankable_field": "",
            },
        )

    def test_rest_custom_lookup_fk(self):
        SlugModel.objects.create(
            code="test1",
            name="Test #1",
        )
        response = self.client.post(
            "/slugrefparents.json",
            {
                "ref_id": "test1",
                "name": "Test FK",
            },
        )
        self.assertTrue(status.is_success(response.status_code), response.data)
        rid = response.data.get("id")
        self.assertEqual(
            response.data,
            {
                "id": rid,
                "label": "Test FK (Test #1)",
                "name": "Test FK",
                "ref_id": "test1",
            },
        )

        response = self.client.post(
            "/slugrefparents.json",
            {
                "ref_id": "test_invalid",
                "name": "Test FK",
            },
        )
        self.assertTrue(
            status.is_client_error(response.status_code), response.data
        )
        self.assertEqual(
            response.data,
            {"ref_id": ["Object with code=test_invalid does not exist."]},
        )

    def test_rest_list_exclude_post(self):
        # Create
        response = self.client.post(
            "/expensivemodels.json",
            {
                "name": "Test",
                "expensive": "SOME_DATA",
            },
        )
        item = response.data
        item_url = "/expensivemodels/{pk}.json".format(pk=item["id"])
        self.assertEqual(item["name"], "Test")
        self.assertEqual(item["expensive"], "SOME_DATA")

        # List view
        response = self.client.get("/expensivemodels.json")
        item = response.data["list"][0]
        self.assertEqual(item["name"], "Test")
        self.assertNotIn("expensive", item)
        self.assertNotIn("more_expensive", item)

        # Detail view
        response = self.client.get(item_url)
        item = response.data
        self.assertEqual(item["name"], "Test")
        self.assertEqual(item["expensive"], "SOME_DATA")

        # Update
        response = self.client.put(
            item_url,
            {
                "name": "Test 2",
                "expensive": "SOME_OTHER_DATA",
            },
        )
        item = response.data
        self.assertEqual(item["name"], "Test 2")
        self.assertEqual(item["expensive"], "SOME_OTHER_DATA")

    def test_rest_virtual_fieldset(self):
        response = self.client.post(
            "/fieldsetmodels.json",
            {
                "general": {
                    "name": "Test",
                    "title": "Dr.",
                },
                "contact": {
                    "address": "123 Main St",
                    "city": "Minneapolis",
                },
            },
            format="json",
        )
        self.assertTrue(status.is_success(response.status_code), response.data)
        obj = FieldsetModel.objects.get(pk=response.data["id"])
        self.assertEqual(obj.name, "Test")
        self.assertEqual(obj.title, "Dr.")
        self.assertEqual(obj.address, "123 Main St")
        self.assertEqual(obj.city, "Minneapolis")

        response = self.client.post(
            "/fieldsetmodels.json",
            {
                "general": {
                    "name": "Test",
                },
                "contact": {
                    "address": "123 Main St",
                    "city": "Minneapolis",
                },
            },
            format="json",
        )
        self.assertTrue(
            status.is_client_error(response.status_code), response.data
        )
        self.assertEqual(
            response.data, {"general": {"title": ["This field is required."]}}
        )

    def test_empty_post(self):
        """
        Empty post should return 400, not 500
        """
        response = self.client.post("/entities.json", data={})
        self.assertTrue(
            status.is_client_error(response.status_code), response.data
        )



from .base import APITestCase
from django.core.management import call_command

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
import json


class CommandTestCase(APITestCase):
    def check_config(self, text):
        data = json.loads(text)
        self.assertIn("pages", data)
        page = list(data["pages"].values())[0]
        self.assertIn("url", page)

    def test_dump_config_json(self):
        f = StringIO()
        call_command("dump_config", stdout=f)
        self.check_config(f.getvalue())

    def test_dump_config_amd(self):
        f = StringIO()
        call_command("dump_config", format="amd", stdout=f)
        text = f.getvalue().strip()
        self.assertTrue(
            text.startswith("define("), "Unexpected start: %s..." % text[:10]
        )
        self.assertTrue(
            text.endswith(");"), "Unexpected end: ...%s" % text[-10:]
        )
        text = text.replace("define(", "")
        text = text.replace(");", "")
        self.check_config(text)



from rest_framework.negotiation import DefaultContentNegotiation


JSON = "application/json"
GEOJSON = "application/geo+json"


class ContentNegotiation(DefaultContentNegotiation):
    """
    Some clients (such as Mapbox GL) might only Accept: application/json
    """

    def filter_renderers(self, renderers, format):
        self._format = format
        return super().filter_renderers(renderers, format)

    def get_accept_list(self, request):
        accepts = super().get_accept_list(request)
        if getattr(self, "_format", None) == "geojson":
            if JSON in accepts and GEOJSON not in accepts:
                accepts.append(GEOJSON)
        return accepts



from rest_framework import serializers
from wq.db.rest.serializers import ModelSerializer
from .models import (
    OneToOneModel,
    ExtraModel,
    Child,
    ChoiceModel,
    DateModel,
    Item,
    ExpensiveModel,
    FieldsetModel,
)


class RootModelSerializer(ModelSerializer):
    extramodels = ModelSerializer.for_model(
        ExtraModel, include_fields="__all__"
    )(many=True)
    onetoonemodel = ModelSerializer.for_model(
        OneToOneModel, include_fields="__all__"
    )()


class ChildSerializer(ModelSerializer):
    class Meta:
        model = Child
        exclude = ("parent",)


class ParentSerializer(ModelSerializer):
    children = ChildSerializer(many=True)


class ItemSerializer(ModelSerializer):
    class Meta:
        wq_field_config = {
            "type": {
                "filter": {
                    "active": [
                        "1",
                        # Allow inactive types when editing existing items
                        "{{#id}}0{{/id}}{{^id}}1{{/id}}",
                    ]
                }
            }
        }


class SlugRefChildSerializer(ModelSerializer):
    class Meta:
        wq_field_config = {"parent": {"filter": {"ref_id": "test"}}}


class ChoiceLabelSerializer(ModelSerializer):
    choice_label = serializers.SerializerMethodField()

    def get_choice_label(self, instance):
        return instance.choice.upper()

    class Meta:
        model = ChoiceModel
        fields = "__all__"


class DateLabelSerializer(ModelSerializer):
    date_label = serializers.SerializerMethodField()

    def get_date_label(self, instance):
        return instance.date.strftime("%B %-d, %Y")

    class Meta:
        model = DateModel
        fields = "__all__"


class ItemLabelSerializer(ModelSerializer):
    label = serializers.SerializerMethodField()
    type_id = serializers.SerializerMethodField()
    type_label = serializers.SerializerMethodField()

    def get_label(self, instance):
        return "%s: %s" % (instance.type.name, instance.name)

    def get_type_id(self, instance):
        return "id-%s" % instance.type_id

    def get_type_label(self, instance):
        return instance.type.name.upper()

    class Meta:
        model = Item
        fields = "__all__"


class ExpensiveSerializer(ModelSerializer):
    class Meta:
        fields = "__all__"
        list_exclude = ("expensive", "more_expensive")
        config_exclude = ("more_expensive",)
        model = ExpensiveModel


class FieldsetSerializer(ModelSerializer):
    status = serializers.SerializerMethodField(
        style={"wq_config": {"disabled": True}}
    )
    is_complete = serializers.SerializerMethodField()

    def get_status(self, instance):
        return "OK"

    def get_is_complete(self, instance):
        return True

    class Meta:
        model = FieldsetModel
        fields = "__all__"
        wq_fieldsets = {
            "general": {
                "label": "General",
                "fields": ["name", "title", "status"],
            },
            "contact": {
                "label": "Contact Information",
                "control": {"appearance": "contact-fieldset"},
                "fields": ["address", "city"],
            },
        }



from django.db import models


class Item(models.Model):
    name = models.CharField(max_length=10)

    def __str__(self):
        return self.name


class TestModel(models.Model):
    name = models.CharField(max_length=10)

    def __str__(self):
        return self.name



from .base import APITestCase
from rest_framework import status
from tests.rest_app.models import (
    RootModel,
    OneToOneModel,
    ForeignKeyModel,
    ExtraModel,
    UserManagedModel,
    Parent,
    ItemType,
    SlugModel,
)
from django.contrib.auth.models import User


class RestTestCase(APITestCase):
    def setUp(self):
        instance = RootModel.objects.create(
            slug="instance",
            description="Test",
        )
        for cls in OneToOneModel, ForeignKeyModel, ExtraModel:
            cls.objects.create(
                root=instance,
            )
        self.user = User.objects.create(username="testuser")
        UserManagedModel.objects.create(id=1, user=self.user)
        parent = Parent.objects.create(name="Test", pk=1)
        parent.children.create(name="Test 1")
        parent.children.create(name="Test 2")
        parent2 = Parent.objects.create(name="Test 2", pk=2)
        parent2.children.create(name="Test 1")
        itype = ItemType.objects.create(name="Test", pk=1)
        itype.item_set.create(name="Test 1")
        itype.item_set.create(name="Test 2")
        SlugModel.objects.create(
            code="test",
            name="Test",
        )

    # Test url="" use case
    def test_rest_list_at_root(self):
        response = self.client.get("/.json")
        self.assertTrue(status.is_success(response.status_code), response.data)
        self.assertTrue(len(response.data["list"]) == 1)

    def test_rest_detail_at_root(self):
        response = self.client.get("/instance.json")
        self.assertTrue(status.is_success(response.status_code), response.data)
        self.assertTrue(response.data["description"] == "Test")

    # Test nested models with foreign keys
    def test_rest_detail_nested_foreignkeys(self):
        response = self.client.get("/instance.json")

        # Include explicitly declared serializers for related fields
        self.assertIn("onetoonemodel", response.data)
        self.assertEqual(
            response.data["onetoonemodel"]["label"],
            "onetoonemodel for instance",
        )
        self.assertIn("extramodels", response.data)
        self.assertEqual(
            response.data["extramodels"][0]["label"], "extramodel for instance"
        )

        # Related fields without explicit serializers will not be included
        self.assertNotIn("extramodel_set", response.data)
        self.assertNotIn("foreignkeymodel_set", response.data)
        self.assertNotIn("foreignkeymodels", response.data)

    def test_rest_filter_by_parent(self):
        response = self.client.get("/parents/1/children.json")
        self.assertIn("list", response.data)
        self.assertEqual(len(response.data["list"]), 2)

        response = self.client.get("/itemtypes/1/items.json")
        self.assertIn("list", response.data)
        self.assertEqual(len(response.data["list"]), 2)

    def test_rest_detail_user_serializer(self):
        response = self.client.get("/usermanagedmodels/1.json")
        self.assertIn("user", response.data)
        self.assertIn("label", response.data["user"])
        self.assertNotIn("password", response.data["user"])

    def test_rest_multi(self):
        lists = ["usermanagedmodels", "items", "children"]
        response = self.client.get("/multi.json?lists=" + ",".join(lists))
        for listurl in lists:
            self.assertIn(listurl, response.data)
            self.assertIn("list", response.data[listurl])
            self.assertGreater(len(response.data[listurl]["list"]), 0)

    def test_rest_custom_lookup(self):
        response = self.client.get("/slugmodels/test.json")
        self.assertTrue(status.is_success(response.status_code), response.data)
        self.assertEqual(response.data["id"], "test")

    def test_rest_default_per_page(self):
        response = self.client.get("/parents.json")
        self.assertTrue(status.is_success(response.status_code), response.data)
        self.assertEqual(response.data["per_page"], 50)

    def test_rest_custom_per_page(self):
        response = self.client.get("/children.json")
        self.assertTrue(status.is_success(response.status_code), response.data)
        self.assertEqual(response.data["per_page"], 100)

    def test_rest_limit(self):
        response = self.client.get("/children.json?limit=10")
        self.assertTrue(status.is_success(response.status_code), response.data)
        self.assertEqual(response.data["per_page"], 10)

    def test_rest_cache_all(self):
        for num in range(2, 101):
            ItemType.objects.create(pk=num, name="Type #%s" % num)

        tests = [
            (100, "/itemtypes.json"),
            (50, "/itemtypes/"),
            (50, "/itemtypes.json?page=1"),
        ]
        for expect_items, url in tests:
            response = self.client.get(url)
            self.assertTrue(
                status.is_success(response.status_code), response.data
            )
            self.assertEqual(
                len(response.data["list"]),
                expect_items,
                "%s should return %s items" % (url, expect_items),
            )
            self.assertEqual(response.data["pages"], 2)
            self.assertEqual(response.data["per_page"], 50)
            self.assertEqual(response.data["count"], 100)

    def test_rest_cache_filter(self):
        other_user = User.objects.create(username="otheruser")
        UserManagedModel.objects.create(id=2, user=other_user)
        UserManagedModel.objects.create(id=3, user=other_user)

        for auth in False, True:
            if auth:
                self.client.force_authenticate(self.user)

            tests = [
                (3, 1 if auth else 0, "/usermanagedmodels.json"),
                (3, 3, "/usermanagedmodels/"),
                (3, 3, "/usermanagedmodels.json?page=1"),
                (1, 1, "/usermanagedmodels.json?user_id=%s" % self.user.pk),
                (2, 2, "/usermanagedmodels.json?user_id=%s" % other_user.pk),
            ]
            for expect_count, expect_items, url in tests:
                response = self.client.get(url)
                self.assertTrue(
                    status.is_success(response.status_code), response.data
                )
                self.assertEqual(
                    len(response.data["list"]),
                    expect_items,
                    "%s should return %s items for %s"
                    % (
                        url,
                        expect_items,
                        "authed user" if auth else "anonymous user",
                    ),
                )
                self.assertEqual(response.data["pages"], 1)
                self.assertEqual(response.data["per_page"], 50)
                self.assertEqual(response.data["count"], expect_count)

    def test_rest_cache_none(self):
        tests = [
            (0, "/items.json"),
            (2, "/items/"),
            (2, "/items.json?page=1"),
        ]
        for expect_items, url in tests:
            response = self.client.get(url)
            self.assertTrue(
                status.is_success(response.status_code), response.data
            )
            self.assertEqual(
                len(response.data["list"]),
                expect_items,
                "%s should return %s items" % (url, expect_items),
            )
            self.assertEqual(response.data["pages"], 1)
            self.assertEqual(response.data["per_page"], 50)
            self.assertEqual(response.data["count"], 2)

    def test_rest_list_head(self):
        response = self.client.head("/")
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.data
        )

    def test_rest_simpleviewset_head(self):
        response = self.client.head("/login")
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.data
        )

    def test_rest_options(self):
        response = self.client.options("/")
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.data
        )

    def test_rest_login_options(self):
        response = self.client.options("/login")
        self.assertEqual(
            response.status_code, status.HTTP_200_OK, response.data
        )



from rest_framework import serializers
from django.core.exceptions import ImproperlyConfigured

try:
    from django.contrib.gis.db import models as model_fields
    from django.contrib.gis.geos import GEOSGeometry
except (OSError, ImproperlyConfigured):
    from django.db import models as model_fields

    GEOSGeometry = None

from django.utils import timezone
from django.core.exceptions import FieldDoesNotExist
from collections import OrderedDict

from django.conf import settings

from rest_framework.utils import model_meta
from html_json_forms.serializers import parse_json_form
from natural_keys.serializers import (
    NaturalKeyModelSerializer,
    NaturalKeySerializer,
)
from drf_writable_nested import WritableNestedModelSerializer
from .exceptions import ImproperlyConfigured


class GeometryField(serializers.Field):
    def to_representation(self, value):
        if value is None:
            return None
        import json

        return json.loads(value.geojson)

    def to_internal_value(self, value):
        import json

        if isinstance(value, dict):
            value = json.dumps(value)

        if not GEOSGeometry:
            raise Exception("Missing GDAL")

        geom = GEOSGeometry(value)
        srid = getattr(settings, "SRID", 4326)

        if "crs" in value and value["crs"].get("type", None) == "name":
            name = value["crs"]["properties"]["name"]
            if name.startswith("urn:ogc:def:crs:EPSG::"):
                geom.srid = int(name.replace("urn:ogc:def:crs:EPSG::", ""))

        if geom.srid is None:
            geom.srid = 4326
        if geom.srid != srid:
            geom.transform(srid)
        return geom


class LookupRelatedField(serializers.SlugRelatedField):
    def __init__(self, router, model, **kwargs):
        self.router = router
        self.model = model
        self._slug_field = kwargs.pop("slug_field", None)
        super(serializers.SlugRelatedField, self).__init__(**kwargs)

    @property
    def slug_field(self):
        if self._slug_field:
            return self._slug_field
        else:
            return self.router.get_lookup_for_model(self.model)

    def use_pk_only_optimization(self):
        if self.slug_field == "pk":
            return True
        if self.parent and self.parent.Meta.model:
            model_field = self.parent.Meta.model._meta.get_field(
                self.source_attrs[-1]
            )
            if model_field.target_field.name == self.slug_field:
                return True
        return False

    def get_attribute(self, instance):
        value = super().get_attribute(instance)
        if self.slug_field != "pk" and isinstance(
            value, serializers.PKOnlyObject
        ):
            setattr(value, self.slug_field, value.pk)
        return value


class LocalDateTimeField(serializers.ReadOnlyField):
    def to_representation(self, value):
        if value is None:
            return None
        if value.tzinfo:
            value = timezone.localtime(value)
        return value.strftime("%Y-%m-%d %I:%M %p")


MODEL_BOOLEAN_FIELDS = (
    model_fields.BooleanField,
    model_fields.NullBooleanField,
)


class BooleanLabelField(serializers.ReadOnlyField):
    def to_representation(self, value):
        if value is None:
            return None
        elif value:
            return "Yes"
        else:
            return "No"


class BlankCharField(serializers.CharField):
    def run_validation(self, data=serializers.empty):
        if self.allow_blank and data is None:
            return ""
        return super(BlankCharField, self).run_validation(data)


def is_empty_or_url(data):
    if not data:
        return True
    if data is serializers.empty:
        return True
    if isinstance(data, str) and data.startswith(("http://", "https://")):
        return True
    return False


class ClearableFileField(serializers.FileField):
    def validate_empty_values(self, data):
        parent = self.parent
        while parent.parent and parent.instance is None:
            parent = parent.parent
        if parent.instance is not None:
            if is_empty_or_url(data):
                raise serializers.SkipField()
            elif data == "__clear__":
                data = None
        elif not data:
            data = None
        return super().validate_empty_values(data)


class ClearableImageField(ClearableFileField, serializers.ImageField):
    pass


def get_choice_list(choices, group=None):
    choice_list = []
    for key, value in choices.items():
        if isinstance(value, dict):
            choice_list += get_choice_list(value, group=key)
        else:
            choice = {
                "name": key,
                "label": value,
            }
            if group:
                choice["group"] = group
            choice_list.append(choice)
    return choice_list


class ListSerializer(serializers.ListSerializer):
    @property
    def type_field(self):
        if not hasattr(self, "_type_field"):
            initial = self.child.get_wq_config().get("initial")
            if (
                initial
                and isinstance(initial, dict)
                and "type_field" in initial
            ):
                self._type_field = initial["type_field"] + "_id"
            else:
                self._type_field = None
        return self._type_field

    def check_empty(self, val):
        return val is None or val == ""

    def skip_empty(self, row):
        if not self.type_field:
            return None
        vals = [
            val
            for key, val in row.items()
            if key != self.type_field and not self.check_empty(val)
        ]
        if not vals:
            return True
        else:
            return False

    def get_value(self, dictionary):
        value = super().get_value(dictionary)
        if isinstance(value, list):
            value = [row for row in value if not self.skip_empty(row)]
        return value


class ModelSerializer(
    NaturalKeyModelSerializer,
    WritableNestedModelSerializer,
):
    add_label_fields = True

    serializer_related_field = LookupRelatedField
    serializer_related_to_field = LookupRelatedField

    xlsform_types = OrderedDict(
        (
            (serializers.ImageField, "image"),
            (ClearableImageField, "image"),
            (serializers.FileField, "file"),
            (ClearableFileField, "file"),
            (serializers.DateField, "date"),
            (serializers.DateTimeField, "dateTime"),
            (serializers.FloatField, "decimal"),
            (GeometryField, "geoshape"),
            (serializers.IntegerField, "int"),
            (serializers.CharField, "string"),
            (serializers.ChoiceField, "select one"),
            (serializers.TimeField, "time"),
        )
    )

    def __init__(self, *args, wq_config=None, **kwargs):
        mapping = self.serializer_field_mapping
        mapping[model_fields.FileField] = ClearableFileField
        mapping[model_fields.ImageField] = ClearableImageField

        for model_field, serializer_field in list(mapping.items()):
            if serializer_field == serializers.CharField:
                mapping[model_field] = BlankCharField

        if GEOSGeometry:
            for field in (
                "Geometry",
                "GeometryCollection",
                "Point",
                "LineString",
                "Polygon",
                "MultiPoint",
                "MultiLineString",
                "MultiPolygon",
            ):
                mapping[getattr(model_fields, field + "Field")] = GeometryField
        if wq_config:
            self.wq_config = wq_config
        super().__init__(*args, **kwargs)
        if getattr(self, "initial_data", None):
            self.initial_data = parse_json_form(self.initial_data)
        self.id_fields = set()

    @classmethod
    def many_init(cls, *args, **kwargs):
        if not hasattr(cls.Meta, "list_serializer_class"):
            cls.Meta.list_serializer_class = ListSerializer
        return super().many_init(*args, **kwargs)

    def create(self, validated_data):
        self.convert_natural_keys(validated_data)
        return WritableNestedModelSerializer.create(self, validated_data)

    def update(self, instance, validated_data):
        self.convert_natural_keys(validated_data)
        return WritableNestedModelSerializer.update(
            self, instance, validated_data
        )

    def get_fields_for_config(self):
        self._for_wq_config = True
        fields = self.get_fields()
        del self._for_wq_config
        overrides = getattr(self.Meta, "wq_field_config", None) or {}

        for name, field in list(fields.items()):
            has_wq_config = False
            if name != "id" and not field.read_only and not field.write_only:
                has_wq_config = True
            elif "wq_config" in field.style:
                has_wq_config = True
            elif name in overrides:
                has_wq_config = True

            if not has_wq_config:
                fields.pop(name)
            elif isinstance(field, serializers.BooleanField):
                fields[name] = serializers.ChoiceField(
                    choices=self.get_boolean_choices(field),
                    required=field.required,
                    label=field.label,
                    help_text=field.help_text,
                )
        return fields

    def get_boolean_choices(self, field):
        return [(True, "Yes"), (False, "No")]

    def get_wq_config(self):
        serializer_fields = self.get_fields_for_config()
        fields = []
        nested_fields = []
        nested_field_dict = {}
        has_geo_fields = False

        overrides = getattr(self.Meta, "wq_field_config", None) or {}
        fieldsets = getattr(self.Meta, "wq_fieldsets", None) or {}

        def get_info(name, field):
            info = self.get_wq_field_info(name, field)
            if info["name"] in overrides:
                info.update(overrides[info["name"]])
            return info

        def pop_info(name):
            fsconf = (
                f"{type(self).__name__}.Meta.wq_fieldsets"
                f" for {self.Meta.model.__name__}"
            )
            if name in serializer_fields:
                field = serializer_fields.pop(name)
                info_name = name
            elif f"{name}_id" in serializer_fields:
                info_name = f"{name}_id"
                field = serializer_fields.pop(info_name)
            else:
                raise ImproperlyConfigured(
                    f'Unknown field "{name}" in {fsconf}'
                )

            is_fk = isinstance(
                field, (serializers.RelatedField, serializers.ManyRelatedField)
            )
            if name == info_name and is_fk:
                raise ImproperlyConfigured(
                    f'Use "{name[:-3]}" rather than "{name}"'
                    f" for ForeignKey in {fsconf}"
                )
            elif name != info_name and not is_fk:
                raise ImproperlyConfigured(
                    f'Use "{info_name}" rather than "{name}"'
                    f" for non-ForeignKey in {fsconf}"
                )
            return get_info(info_name, field)

        for name, conf in fieldsets.items():
            conf = conf.copy()
            conf.setdefault("name", name)
            conf.setdefault("type", "group")
            conf.setdefault("label", name)
            conf["children"] = [
                pop_info(name) for name in conf.pop("fields", [])
            ]
            nested_fields.append(conf)
            nested_field_dict[name] = conf

        for name, field in serializer_fields.items():
            info = get_info(name, field)

            if info["type"].startswith("geo") or info["name"] == "latitude":
                has_geo_fields = True

            if info["type"] == "repeat":
                if name in nested_field_dict:
                    conf = nested_field_dict[name]
                    for key, val in info.items():
                        if key in ("type", "children"):
                            conf[key] = val
                        else:
                            conf.setdefault(key, val)
                else:
                    nested_fields.append(info)
                    nested_field_dict[name] = info
            else:
                fields.append(info)

        config = getattr(self.Meta, "wq_config", {}).copy()
        config.update(getattr(self, "wq_config", {}))
        config["form"] = fields + nested_fields

        meta = self.Meta.model._meta
        config.setdefault("name", meta.model_name)
        config.setdefault("verbose_name", meta.verbose_name)
        config.setdefault("verbose_name_plural", meta.verbose_name_plural)
        if meta.ordering:
            config.setdefault("ordering", meta.ordering)

        if has_geo_fields:
            config.setdefault("map", True)

        label_template = getattr(self.Meta.model, "wq_label_template", None)
        if label_template:
            config.setdefault("label_template", label_template)
        return config

    def get_wq_field_info(self, name, field, model=None):
        if isinstance(field, NaturalKeySerializer):
            children = [
                self.get_wq_field_info(n, f, model=field.Meta.model)
                for n, f in field.get_fields().items()
            ]
            if len(children) == 1:
                info = children[0]
                info["name"] = name + "[%s]" % info["name"]

                fk = self.get_wq_foreignkey_info(field.Meta.model)
                if fk:
                    info["wq:ForeignKey"] = fk
                    info["type"] = "select one"
            else:
                info = {
                    "name": name,
                    "type": "group",
                    "bind": {"required": True},
                    "children": children,
                }
            info["label"] = field.label or name.replace("_", " ").title()
            return info

        info = {
            "name": name,
            "label": field.label or name.replace("_", " ").title(),
        }
        if field.required:
            info.setdefault("bind", {})
            info["bind"]["required"] = True

        if field.help_text:
            info["hint"] = field.help_text

        if isinstance(field, serializers.ChoiceField):
            info["choices"] = get_choice_list(field.grouped_choices)
        elif getattr(field, "max_length", None):
            if not isinstance(field, serializers.FileField):
                info["wq:length"] = field.max_length

        if isinstance(field, serializers.BaseSerializer):
            # Nested model with a foreign key to this one
            if isinstance(field, serializers.ListSerializer):
                info["type"] = "repeat"
                serializer = field.child
            else:
                info["type"] = "group"
                serializer = field

            serializer.context["router"] = self.router
            if hasattr(serializer, "get_wq_config"):
                child_config = serializer.get_wq_config()
            else:
                child_config = {"form": []}

            for key in child_config:
                if key in (
                    "name",
                    "verbose_name",
                    "verbose_name_plural",
                    "url",
                    "postsave",
                    "label_template",
                ):
                    continue
                elif key == "form":
                    info_key = "children"
                else:
                    info_key = key
                info[info_key] = child_config[key]

        elif isinstance(field, serializers.RelatedField):
            # Foreign key to a parent model
            info["type"] = "select one"
            info["name"] = info["name"].replace("_id", "")
            info["label"] = info["label"].replace(" Id", "")
            if hasattr(field, "queryset"):
                fk = self.get_wq_foreignkey_info(field.queryset.model)
                if fk:
                    model = model or self.Meta.model
                    source = model._meta.get_field(name)
                    info["wq:ForeignKey"] = fk
                    info[
                        "wq:related_name"
                    ] = source.remote_field.get_accessor_name()

        elif isinstance(field, serializers.ManyRelatedField):
            # ManyToMany field to related model
            info["type"] = "select"
            info["name"] = info["name"].replace("_id", "")
            info["label"] = info["label"].replace(" Id", "")
            if field.child_relation.queryset:
                fk = self.get_wq_foreignkey_info(
                    field.child_relation.queryset.model
                )
                info["wq:ForeignKey"] = fk

        if "type" not in info:
            info["type"] = "string"

            for field_type, xlsform_type in self.xlsform_types.items():
                if isinstance(field, field_type):
                    info["type"] = xlsform_type
                    break

            if info["type"] == "geoshape":
                model = model or self.Meta.model
                source = model._meta.get_field(name)
                if info["type"] == "geoshape":
                    geom_type = getattr(source, "geom_type", None)
                    if geom_type == "POINT":
                        info["type"] = "geopoint"
                    elif geom_type == "LINESTRING":
                        info["type"] = "geotrace"

            if info["type"] == "string":
                model = model or self.Meta.model
                try:
                    source = model._meta.get_field(name)
                except FieldDoesNotExist:
                    pass
                else:
                    if source.get_internal_type() == "TextField":
                        info["type"] = "text"

        if "wq_config" in field.style:
            info.update(field.style["wq_config"])

        field_config = getattr(self.Meta, "wq_field_config", {})
        if name in field_config:
            info.update(field_config[name])

        return info

    def get_wq_foreignkey_info(self, model):
        if not self.router or not self.router.model_is_registered(model):
            return None
        else:
            serializer = self.router.get_serializer_for_model(model)
            model_conf = serializer().get_wq_config()
            return model_conf["name"]

    def to_representation(self, obj):
        data = super().to_representation(obj)
        if getattr(self.Meta, "wq_fieldsets", None):
            return self.to_virtual_fieldsets(data)
        else:
            return data

    def to_virtual_fieldsets(self, data):
        for name, conf in self.Meta.wq_fieldsets.items():
            if name == "":
                continue
            for field in conf.get("fields") or []:
                if field in self.id_fields:
                    field = f"{field}_id"
                if field not in data:
                    continue
                data.setdefault(name, {})
                data[name][field] = data.pop(field)
        return data

    def to_internal_value(self, data):
        if not data or not getattr(self.Meta, "wq_fieldsets", None):
            return super().to_internal_value(data)

        for name, conf in self.Meta.wq_fieldsets.items():
            if name == "" or not conf.get("fields"):
                continue
            fs_data = data.pop(name, None)
            if isinstance(fs_data, dict):
                data.update(fs_data)

        try:
            return super().to_internal_value(data)
        except serializers.ValidationError as e:
            if isinstance(e.detail, dict):
                raise serializers.ValidationError(
                    self.to_virtual_fieldsets(e.detail)
                )
            else:
                raise

    @property
    def router(self):
        if "view" not in self.context and "router" not in self.context:
            return None
        if "view" in self.context:
            return self.context["view"].router
        else:
            return self.context["router"]

    @property
    def is_detail(self):
        if getattr(self.Meta, "depth", 0) > 0:
            return True
        view = self.context.get("view", None)
        return view and view.action != "list"

    @property
    def is_edit(self):
        view = self.context.get("view", None)
        return view and view.action == "edit"

    @property
    def is_geojson(self):
        request = self.context.get("request", None)
        if not request:
            return
        if request.accepted_renderer.format == "geojson":
            return True
        return False

    @property
    def is_html(self):
        request = self.context.get("request", None)
        if not request:
            return
        if request.accepted_renderer.format == "html":
            return True
        return False

    @property
    def is_config(self):
        return getattr(self, "_for_wq_config", False)

    def get_fields(self, *args, **kwargs):
        fields = super().get_fields(*args, **kwargs)
        fields = self.update_id_fields(fields)
        fields.update(self.get_label_fields(fields))
        fields.update(self.get_nested_arrays(fields))
        exclude = set()

        def get_exclude(meta_name):
            return set(getattr(self.Meta, meta_name, []))

        if not self.is_detail and not self.is_config:
            exclude |= get_exclude("list_exclude")
            if self.is_html:
                exclude |= get_exclude("html_list_exclude")

        if self.is_config:
            exclude |= get_exclude("config_exclude")

        for field in list(exclude):
            fields.pop(field, None)

        return fields

    def update_id_fields(self, fields):
        if "id" not in self._declared_fields:
            lookup = getattr(self.Meta, "wq_config", {}).get("lookup", None)
            if lookup and lookup != "id":
                fields["id"] = serializers.ReadOnlyField(source=lookup)

        id_fields = {}
        info = model_meta.get_field_info(self.Meta.model)

        # In list views, remove [fieldname] as an attribute in favor of
        # [fieldname]_id.
        extra_kwargs = self.get_extra_kwargs()
        for name, field in info.forward_relations.items():
            include = getattr(self.Meta, "fields", [])
            if include == "__all__":
                include = []
            exclude = getattr(self.Meta, "exclude", [])
            if name in exclude or (include and name not in include):
                fields.pop(name, None)
                continue

            default_field = fields[name]
            auto_related_field = (
                serializers.BaseSerializer,
                serializers.ManyRelatedField,
                serializers.SlugRelatedField,
            )
            if not isinstance(default_field, auto_related_field):
                continue
            if isinstance(default_field, NaturalKeySerializer):
                continue

            if name + "_id" not in fields:
                id_field, id_field_kwargs = self.build_relational_field(
                    name, field
                )
                extra_field_kwargs = extra_kwargs.get(name, {})
                extra_field_kwargs.update(extra_kwargs.get(name + "_id", {}))
                self.include_extra_kwargs(
                    id_field_kwargs,
                    extra_field_kwargs,
                )
                id_field_kwargs["source"] = name
                id_fields[name] = id_field(**id_field_kwargs)

            if name in fields:
                # Update/remove DRF default foreign key field (w/o _id)
                if self.is_detail and isinstance(
                    default_field, serializers.BaseSerializer
                ):
                    # Nested representation, keep for detail template context
                    fields[name].read_only = True
                else:
                    # Otherwise we don't need this field
                    del fields[name]

        if not id_fields:
            return fields

        self.id_fields.update(id_fields)

        # Ensure _id fields show up in the same order they appear on model
        updated_fields = OrderedDict()
        for field in self.Meta.model._meta.fields:
            name = field.name
            if name in fields:
                updated_fields[name] = fields.pop(name)
            if name in id_fields:
                updated_fields[name + "_id"] = id_fields.pop(name)

        if fields:
            updated_fields.update(fields)

        if id_fields:
            for name in id_fields:
                updated_fields[name + "_id"] = id_fields[name]

        return updated_fields

    def get_label_fields(self, default_fields):
        if not self.add_label_fields:
            return {}
        fields = {}

        exclude = getattr(self.Meta, "exclude", [])
        if "label" not in exclude and "label" not in default_fields:
            fields["label"] = serializers.ReadOnlyField(source="__str__")

        info = model_meta.get_field_info(self.Meta.model)

        # Add labels for dates and fields with choices
        for name, field in info.fields.items():
            if name in getattr(self.Meta, "exclude", []):
                continue
            if name + "_label" in default_fields:
                continue
            if isinstance(field, model_fields.DateTimeField):
                fields[name + "_label"] = LocalDateTimeField(source=name)
            if field.choices:
                fields[name + "_label"] = serializers.ReadOnlyField(
                    source="get_%s_display" % name
                )
            if isinstance(field, MODEL_BOOLEAN_FIELDS):
                fields[name + "_label"] = BooleanLabelField(
                    source=name,
                )

        return fields

    def get_nested_arrays(self, fields):
        model_class = self.Meta.model
        nested = {}
        for array_model in getattr(self.Meta, "wq_nested_arrays", None) or []:
            fk_field = None
            for field in array_model._meta.fields:
                if field.related_model == model_class:
                    fk_field = field
                    break
            fk_desc = (
                f"ForeignKey from {array_model.__name__}"
                f" to {model_class.__name__}"
            )
            if not fk_field:
                raise ImproperlyConfigured(f"No {fk_desc}")

            rel_name = fk_field.remote_field.related_name
            if not rel_name or rel_name.startswith("+"):
                raise ImproperlyConfigured(f"No related_name for {fk_desc}")

            if rel_name in fields:
                continue

            base = self.get_nested_array_serializer(array_model)
            base_fields = getattr(base.Meta, "fields", None) or []
            base_exclude = getattr(base.Meta, "exclude", None) or []

            class NestedSerializer(base):
                class Meta(base.Meta):
                    model = array_model
                    if not base_fields or base_fields == "__all__":
                        fields = None
                        if fk_field.name not in base_exclude:
                            exclude = [*base_exclude, fk_field.name]
                    else:
                        fields = [f for f in base_fields if f != fk_field.name]

            nested[rel_name] = NestedSerializer(many=True)
        return nested

    def get_nested_array_serializer(self, array_model):
        if self.router:
            return self.router._serializers.get(array_model, ModelSerializer)
        else:
            return ModelSerializer

    @classmethod
    def for_model(
        cls,
        model_class,
        include_fields=None,
        nested_arrays=None,
        serializer_depth=None,
    ):
        class Serializer(cls):
            class Meta(getattr(cls, "Meta", object)):
                model = model_class
                if include_fields:
                    fields = include_fields
                if nested_arrays:
                    wq_nested_arrays = nested_arrays
                if serializer_depth is not None:
                    depth = serializer_depth

        return Serializer

    def build_relational_field(self, field_name, relation_info):
        if isinstance(relation_info.related_model, str):
            raise ImproperlyConfigured(
                "%s could not be resolved to a model class"
                % relation_info.related_model,
            )

        field_class, field_kwargs = super(
            ModelSerializer, self
        ).build_relational_field(field_name, relation_info)
        if field_class == LookupRelatedField:
            field_kwargs["router"] = self.router
            field_kwargs["model"] = relation_info.related_model
        return field_class, field_kwargs

    def build_nested_field(self, field_name, relation_info, nested_depth):
        field_class, field_kwargs = super(
            ModelSerializer, self
        ).build_nested_field(field_name, relation_info, nested_depth)
        field_kwargs["required"] = False
        if self.router and not issubclass(field_class, NaturalKeySerializer):
            field_class = self.router.get_serializer_for_model(
                relation_info.related_model, nested_depth
            )
        return field_class, field_kwargs

    class Meta:
        wq_config = {}
        wq_field_config = {}
        wq_fieldsets = None
        wq_nested_arrays = None
        list_serializer_class = ListSerializer



from wq.db import rest
from .models import NaturalKeyParent, ModelWithNaturalKey


rest.router.register(NaturalKeyParent, fields="__all__")
rest.router.register(ModelWithNaturalKey, fields="__all__")



from rest_framework import serializers
from wq.db.rest.serializers import ModelSerializer


class LoginSerializer(serializers.Serializer):
    username = serializers.CharField()
    password = serializers.CharField(style={"input_type": "password"})


class UserSerializer(ModelSerializer):
    class Meta:
        exclude = (
            "id",
            "password",
            "user_permissions",
            "groups",
        )
        read_only_fields = ("last_login", "date_joined")



from rest_framework.renderers import TemplateHTMLRenderer, JSONRenderer
from django.conf import settings
import re


APP_TEMPLATES = {}
SRID = 4326


def load_app_template(template_name):
    with open(template_name) as f:
        template = f.read()
    template = re.sub(
        "<title>(.+)</title>", "<title>{{title}}</title>", template
    )
    if "{{" in template:
        return template, True
    else:
        return template, False


def get_title(data, request):
    title = None
    if isinstance(data, dict):
        title = data.get("label")

    return title or settings.PROJECT_NAME


def render_app(template_name, data, request):
    if template_name not in APP_TEMPLATES:
        APP_TEMPLATES[template_name] = load_app_template(template_name)

    template, has_title = APP_TEMPLATES[template_name]
    if has_title:
        from wq.db.rest import router

        return template.replace("{{title}}", get_title(data, request)).replace(
            "{{base_url}}", router.get_base_url()
        )
    else:
        return template


class HTMLRenderer(TemplateHTMLRenderer):
    def render(self, data, accepted_media_type=None, renderer_context=None):
        if getattr(settings, "WQ_APP_TEMPLATE", None):
            return render_app(
                settings.WQ_APP_TEMPLATE,
                data,
                (renderer_context or {}).get("request"),
            )
        return super().render(data, accepted_media_type, renderer_context)


class ESMRenderer(JSONRenderer):
    media_type = "application/javascript"
    format = "js"

    def render(self, data, accepted_media_type=None, renderer_context=None):
        data = super().render(data, accepted_media_type, renderer_context)
        return b"const config = " + data + b";\nexport default config;"


class GeoJSONRenderer(JSONRenderer):
    media_type = "application/geo+json"
    format = "geojson"

    def render(self, data, *args, **kwargs):
        if isinstance(data, list):
            features, simple = self.render_features(data)
            data = {"type": "FeatureCollection", "features": features}
        elif "list" in data and isinstance(data["list"], list):
            features, simple = self.render_features(data["list"])
            data["type"] = "FeatureCollection"
            data["features"] = features
            del data["list"]

        else:
            data, simple = self.render_feature(data)

        if not simple and getattr(settings, "SRID", SRID) != SRID:
            data["crs"] = {
                "type": "name",
                "properties": {
                    "name": "urn:ogc:def:crs:EPSG::%s" % settings.SRID
                },
            }
        return super().render(data, *args, **kwargs)

    def render_feature(self, obj):
        feature = {"type": "Feature", "properties": obj}
        simple = False
        if "id" in obj:
            feature["id"] = obj["id"]
            del obj["id"]

        if "latitude" in obj and "longitude" in obj:
            feature["geometry"] = {
                "type": "Point",
                "coordinates": [obj["longitude"], obj["latitude"]],
            }
            del obj["latitude"]
            del obj["longitude"]
            simple = True

        else:
            for key, val in list(obj.items()):
                if (
                    isinstance(val, dict)
                    and "type" in val
                    and ("coordinates" in val or "geometries" in val)
                ):
                    feature["geometry"] = val
                    del obj[key]

        if "features" in obj:
            feature["features"] = obj["features"]
            feature["type"] = "FeatureCollection"
            del obj["features"]

        return feature, simple

    def render_features(self, objs):
        features = []
        has_simple = False
        for obj in objs:
            feature, simple = self.render_feature(obj)
            if simple:
                has_simple = True
                if feature["geometry"]["coordinates"][0] is not None:
                    features.append(feature)
            else:
                features.append(feature)
        return features, has_simple



from rest_framework.permissions import BasePermission


class ModelPermissions(BasePermission):
    METHOD_PERM = {
        "GET": "view",
        "HEAD": "view",
        "OPTIONS": "view",
        "POST": "add",
        "PUT": "change",
        "PATCH": "change",
        "DELETE": "delete",
    }

    def has_permission(self, request, view):
        if getattr(view, "model", None) is None:
            return True
        user = request.user
        method_perm = self.METHOD_PERM.get(
            request.method, self.METHOD_PERM["GET"]
        )
        return has_perm(user, view.model, method_perm)


def has_perm(user, model, method_perm):
    app_label = model._meta.app_label
    model_name = model._meta.model_name
    perm = f"{app_label}.{method_perm}_{model_name}"

    if method_perm == "view":
        return True
    elif user.is_authenticated:
        return user.has_perm(perm)
    else:
        from django.conf import settings

        return perm in getattr(settings, "ANONYMOUS_PERMISSIONS", {})



from django.utils.encoding import force_str
from django.urls import re_path

from django.conf import settings
from rest_framework.routers import DefaultRouter, Route
from rest_framework.urlpatterns import format_suffix_patterns
from rest_framework.settings import api_settings
from rest_framework.response import Response

from .permissions import has_perm
from .views import SimpleViewSet, ModelViewSet
from .renderers import JSONRenderer, ESMRenderer
from .exceptions import ImproperlyConfigured


class ModelRouter(DefaultRouter):
    _models = set()
    _serializers = {}
    _nested_arrays = {}
    _fields = {}
    _querysets = {}
    _filters = {}
    _cache_filters = {}
    _viewsets = {}
    _extra_pages = {}
    _config = {}
    _page_names = {}
    _page_models = {}
    _url_models = {}
    _extra_config = {}

    include_root_view = False
    include_config_view = True
    include_multi_view = True

    default_serializer_class = None

    def __init__(self, trailing_slash=False):
        # Add trailing slash for HTML list views
        self.routes.append(
            Route(
                url=r"^{prefix}/$",
                mapping={"get": "list", "post": "create"},
                name="{basename}-list",
                detail=False,
                initkwargs={"suffix": "List"},
            )
        )
        super(ModelRouter, self).__init__(trailing_slash=trailing_slash)

    def register(self, model, *args, **kwargs):
        if isinstance(model, type) and not args:
            self.register_model(model, **kwargs)
        else:
            super().register(model, *args, **kwargs)

    def register_model(
        self,
        model,
        viewset=None,
        serializer=None,
        nested_arrays=None,
        fields=None,
        queryset=None,
        filter=None,
        cache_filter=None,
        **kwargs,
    ):
        if isinstance(model, str) and "." in model:
            from django.db.models import get_model

            model = get_model(*model.split("."))

        if viewset:
            self.register_viewset(model, viewset)
        if queryset is not None:
            self.register_queryset(model, queryset)
        if filter:
            self.register_filter(model, filter)
        if cache_filter:
            self.register_cache_filter(model, cache_filter)
            kwargs.setdefault("cache", "filter")

        if "name" not in kwargs:
            kwargs["name"] = model._meta.model_name
        if "url" not in kwargs:
            url = force_str(model._meta.verbose_name_plural)
            kwargs["url"] = url.replace(" ", "")

        other_model = self._page_models.get(kwargs["name"], None)
        if other_model:
            raise ImproperlyConfigured(
                "Could not register %s: "
                "the name '%s' was already registered for %s"
                % (model, kwargs["name"], other_model)
            )

        other_model = self._url_models.get(kwargs["url"], None)
        if other_model:
            raise ImproperlyConfigured(
                "Could not register %s: "
                "the url '%s' was already registered for %s"
                % (model, kwargs["url"], other_model)
            )

        self.register_config(model, kwargs)

        if serializer:
            self.register_serializer(model, serializer)

        if nested_arrays:
            if not isinstance(nested_arrays, (list, tuple)):
                nested_arrays = [nested_arrays]
            self._nested_arrays[model] = nested_arrays

        if fields:
            self.register_fields(model, fields)

        self._models.add(model)
        self._page_names[model] = kwargs["name"]
        self._page_models[kwargs["name"]] = model
        self._url_models[kwargs["url"]] = model

    def register_viewset(self, model, viewset):
        self._viewsets[model] = viewset

    def register_serializer(self, model, serializer):
        self._serializers[model] = serializer
        self._base_config = None

    def register_fields(self, model, fields):
        self._fields[model] = fields

    def register_queryset(self, model, queryset):
        self._querysets[model] = queryset

    def register_filter(self, model, filter):
        self._filters[model] = filter

    def register_cache_filter(self, model, cache_filter):
        self._cache_filters[model] = cache_filter

    def register_config(self, model, config):
        for key in ("partial", "reversed", "max_local_pages"):
            if key in config:
                raise ImproperlyConfigured(
                    '"%s" is deprecated in favor of "cache"' % key
                )
        self._config[model] = config
        self._base_config = None

    def update_config(self, model, **kwargs):
        if model not in self._config:
            raise RuntimeError("%s must be registered first" % model)
        for key in ("partial", "reversed", "max_local_pages"):
            if key in kwargs:
                raise ImproperlyConfigured(
                    '"%s" is deprecated in favor of "cache"' % key
                )
        self._config[model].update(kwargs)
        self._base_config = None

    def set_extra_config(self, **extra):
        self._extra_config.update(extra)
        self._base_config = None

    def get_default_serializer_class(self, model_class=None):
        if not self.default_serializer_class:
            from .serializers import ModelSerializer

            self.default_serializer_class = ModelSerializer
        return self.default_serializer_class

    def get_class(self, classes, model_class, default=lambda model: None):
        if model_class in classes:
            return classes[model_class]
        else:
            # Make sure we're not dealing with a proxy
            real_model = model_class._meta.concrete_model
            if real_model in classes:
                return classes[real_model]
            else:
                return default(real_model)

    def get_serializer_for_model(self, model_class, serializer_depth=None):
        serializer = self.get_class(
            self._serializers, model_class, self.get_default_serializer_class
        )

        meta = getattr(serializer, "Meta", object)
        if not getattr(meta, "model", None):
            if getattr(meta, "fields", None) or getattr(meta, "exclude", None):
                include_fields = None
            else:
                include_fields = self._fields.get(model_class, None)
                if not include_fields:
                    raise ImproperlyConfigured(
                        "No serializer fields defined for %s" % model_class
                    )
            serializer = serializer.for_model(
                model_class,
                include_fields=include_fields,
                serializer_depth=serializer_depth,
                nested_arrays=self._nested_arrays.get(model_class),
            )

        meta = getattr(serializer, "Meta", object)
        conf = self._config.get(model_class, None)
        if not conf:
            conf = self._config.get(model_class._meta.concrete_model, None)

        if conf:
            wq_conf = getattr(meta, "wq_config", {}).copy()
            wq_conf.update(conf)
            meta.wq_config = wq_conf
            serializer.Meta = meta

        return serializer

    def serialize(self, obj, many=False, depth=None, request=None):
        if many:
            # assume obj is a queryset
            model = obj.model
            if depth is None:
                depth = 0
        else:
            model = obj.__class__
            if depth is None:
                depth = 1
        serializer = self.get_serializer_for_model(model, depth)
        context = {"router": self}
        if request:
            context["request"] = request
        return serializer(obj, many=many, context=context).data

    def get_object_id(self, instance):
        return getattr(instance, self.get_lookup_for_model(instance))

    def get_by_identifier(self, queryset, ident):
        if hasattr(queryset, "get_by_identifier"):
            return queryset.get_by_identifier(ident)
        else:
            lookup = self.get_lookup_for_model(queryset.model)
            return queryset.get(**{lookup: ident})

    def get_paginate_by_for_model(self, model_class):
        config = self.get_model_config(model_class) or {}
        paginate_by = config.get("per_page", None)
        if paginate_by:
            return paginate_by
        return api_settings.PAGE_SIZE

    def paginate(self, model, page_num, request):
        # FIXME: should copy() before modifying but doing so causes recursion
        request.GET = {
            "page": page_num,
        }
        view = self.get_viewset_for_model(model).as_view(
            actions={"get": "list"},
        )
        return view(request._request).data

    def get_queryset_for_model(self, model, request=None):
        if model in self._page_models:
            model = self._page_models[model]
        if model in self._querysets:
            qs = self._querysets[model]
        else:
            qs = model.objects.all()
        if request and model in self._filters:
            qs = self._filters[model](qs, request)
        return qs

    def get_cache_filter_for_model(self, model):
        return self._cache_filters.get(model, lambda qs, req: qs.none())

    def get_lookup_for_model(self, model_class):
        config = self.get_model_config(model_class) or {}
        return config.get("lookup", "pk")

    def get_viewset_for_model(self, model_class):
        if model_class in self._page_models:
            model_class = self._page_models[model_class]
        viewset = self.get_class(
            self._viewsets, model_class, lambda d: ModelViewSet
        )
        lookup = self.get_lookup_for_model(model_class)

        per_page = self.get_paginate_by_for_model(model_class)
        if per_page != api_settings.PAGE_SIZE:

            class CustomPagination(api_settings.DEFAULT_PAGINATION_CLASS):
                page_size = per_page

        else:
            CustomPagination = None

        class ViewSet(viewset):
            model = model_class
            router = self
            lookup_field = lookup

            if CustomPagination:
                pagination_class = CustomPagination

        return ViewSet

    def get_base_url(self):
        return getattr(settings, "WQ_BASE_URL", "")

    _base_config = None

    @property
    def base_config(self):
        if self._base_config:
            return self._base_config

        from django.contrib.auth.models import AnonymousUser

        user = AnonymousUser()
        pages = {}
        for page in self._extra_pages:
            conf, view = self.get_page(page)
            pages[page] = conf
        for model in self._models:
            if not has_perm(user, model, "view"):
                continue

            info = self._config[model].copy()
            info["list"] = True
            for perm in ("add", "change", "delete"):
                if has_perm(user, model, perm):
                    info["can_" + perm] = True

            serializer = self.get_serializer_for_model(model)
            if not hasattr(serializer, "get_wq_config"):
                raise Exception(model)
            conf = serializer(context={"router": self}).get_wq_config()
            for key in conf:
                if key not in info:
                    info[key] = conf[key]

            pages[info["name"]] = info

        site_title = getattr(settings, "PROJECT_NAME", None)
        base_url = self.get_base_url()

        if not any(conf.get("url") == "" for conf in pages.values()):
            pages.setdefault(
                "index",
                {
                    "url": "",
                    "name": "index",
                    "show_in_index": False,
                    "verbose_name": site_title,
                },
            )

        self._base_config = {
            "pages": pages,
            "site_title": site_title,
            "router": {"base_url": base_url},
            "store": {
                "service": base_url,
                "defaults": {"format": "json"},
            },
        }
        if getattr(settings, "WQ_CONFIG", None):
            self._base_config.update(settings.WQ_CONFIG)
        if settings.DEBUG:
            self._base_config["debug"] = True
        self._base_config.update(self._extra_config)
        return self._base_config

    @property
    def config(self):
        return self.base_config

    def get_user_config(self, user):
        # Add user-specific permissions to configuration
        config = {"pages": {}}
        for page, info in self.config["pages"].items():
            if not info.get("list", False):
                continue
            model = self._page_models[page]
            perms = {}
            for perm in ("add", "change", "delete"):
                if has_perm(user, model, perm):
                    perms["can_" + perm] = True
            config["pages"][page] = perms

        return config

    def add_page(self, name, config, view=None):
        if view is None:

            class PageView(SimpleViewSet):
                template_name = name + ".html"

                def list(self, request, *args, **kwargs):
                    return Response(config)

            view = PageView
        if "name" not in config:
            config["name"] = name
        if "url" not in config:
            config["url"] = name
        self._extra_pages[name] = config, view
        self._base_config = None

    def get_page(self, page):
        return self._extra_pages[page]

    def get_page_config(self, name):
        return self.config["pages"].get(name, None)

    def get_model_config(self, model):
        if not isinstance(model, type):
            model = type(model)
        if model in self._page_models:
            model = self._page_models[model]

        # First, check models registered with API
        if model in self._page_names:
            return self.config["pages"][self._page_names[model]]

        # Then check config cache directly (in case model was configured but
        # not fully registered as part of API)
        if model in self._config:
            return self._config[model]
        return None

    def model_is_registered(self, model):
        return model in self._models

    def get_foreign_keys(self, model):
        rel_models = {}
        for f in model._meta.fields:
            rel = f.remote_field
            if (
                rel is not None
                and type(rel).__name__ == "ManyToOneRel"
                and self.model_is_registered(rel.model)
            ):
                rel_models.setdefault(rel.model, [])
                rel_models[rel.model].append(f.name)
        return rel_models

    def get_config_view(self):
        class ConfigView(SimpleViewSet):
            renderer_classes = [JSONRenderer, ESMRenderer]

            def list(this, request, *args, **kwargs):
                return Response(self.config)

        return ConfigView

    def get_index_view(self):
        class IndexView(SimpleViewSet):
            def list(this, request, *args, **kwargs):
                return Response({})

        return IndexView

    def get_multi_view(self):
        class MultipleListView(SimpleViewSet):
            def list(this, request, *args, **kwargs):
                urls = request.GET.get("lists", "").split(",")
                return self.get_multi(request, urls)

        return MultipleListView

    def get_multi(self, request, urls):
        conf_by_url = {
            conf["url"]: (page, conf)
            for page, conf in self.config["pages"].items()
        }
        result = {}
        for listurl in urls:
            if listurl not in conf_by_url:
                continue
            page, conf = conf_by_url[listurl]
            model = self._page_models[page]
            result[listurl] = self.paginate(model, 1, request)
        return Response(result)

    def get_urls(self):
        # Register viewsets with DefaultRouter just before returning urls

        # The root viewset (with a url of "") should be registered last
        root = {}

        def register(config, viewset):
            if config["url"] == "":
                root["view"] = viewset
                root["name"] = config["name"]
            else:
                self.register(config["url"], viewset, config["name"])

        # List (model) views
        for model in self._models:
            viewset = self.get_viewset_for_model(model)
            config = self._config[model]
            register(config, viewset)

        # Extra/custom viewsets
        for name, (config, viewset) in self._extra_pages.items():
            if not config.get("external"):
                register(config, viewset)

        if self.include_config_view:
            # /config.js
            self.register("config", self.get_config_view(), "config")

        if self.include_multi_view:
            # /multi.json
            self.register("multi", self.get_multi_view(), "multi")

        if not root:
            # default index
            self.register("", self.get_index_view(), "index")

        urls = super(ModelRouter, self).get_urls()

        if root:
            # / - Skip registration and directly generate custom URLs
            urls.extend(self.get_root_urls(root["view"], root["name"]))

        return urls

    def get_root_urls(self, viewset, name):
        lookup = self.get_lookup_regex(viewset)
        routes = self.get_routes(viewset)
        urls = []
        for route in routes:
            pattern = route.url.format(
                prefix="",
                lookup=lookup,
                trailing_slash=self.trailing_slash,
            )

            # Remove leading slash from detail view URLs
            pattern = pattern.replace("/", "", 1)

            mapping = self.get_method_map(viewset, route.mapping)
            if not mapping:
                continue
            view = viewset.as_view(mapping, **route.initkwargs)
            urls.append(re_path(pattern, view, name=name))

        return format_suffix_patterns(urls)

    def get_routes(self, viewset):
        routes = super(ModelRouter, self).get_routes(viewset)
        model = getattr(viewset, "model", None)
        if not model:
            return routes

        # Custom routes

        # Re-register list view, with an additional keyword to filter this
        # model by parent models (e.g. foreign keys)

        # /[parentmodel_url]/[foreignkey_value]/[model_url]
        for rel_model, fields in self.get_foreign_keys(model).items():
            if len(fields) > 1:
                # Multiple foreign keys to same parent model; can't
                # automatically determine which one to use
                continue
            purlbase = self.get_model_config(rel_model)["url"]
            if purlbase != "":
                purlbase += "/"
            routes.append(
                Route(
                    url=(
                        "^"
                        + purlbase
                        + r"(?P<"
                        + fields[0]
                        + r">[^\/\?]+)/{prefix}{trailing_slash}$"
                    ),
                    mapping={"get": "list"},
                    name="{basename}-for-%s" % fields[0],
                    detail=False,
                    initkwargs={"suffix": "List"},
                )
            )

        if hasattr(viewset, "extra_routes"):
            routes += viewset.extra_routes()

        return routes

    @property
    def version(self):
        if not hasattr(self, "_version"):
            vtxt = getattr(settings, "VERSION_TXT", None)
            if vtxt is None:
                self._version = None
            else:
                vfile = open(vtxt, "r")
                self._version = vfile.read()
                vfile.close()
        return self._version

    @property
    def urls(self):
        urls = super(ModelRouter, self).urls
        return urls, "wq", "wq"


# Default router instance, c.f. django.contrib.admin.sites.site
router = ModelRouter()



from django.utils.module_loading import autodiscover_modules
from .routers import ModelRouter, router
from .management.commands.dump_config import dump_config
from django.conf import settings
from rest_framework.serializers import SerializerMetaclass


__all__ = (
    "ModelSerializer",
    "ModelRouter",
    "router",
    "register",
    "autodiscover",
)


class ModelSerializer(metaclass=SerializerMetaclass):
    def __init_subclass__(cls):
        from .serializers import ModelSerializer

        cls.__bases__ = (ModelSerializer,)

    def __instancecheck__(self, instance):
        from .serializers import ModelSerializer

        return isinstance(instance, ModelSerializer)

    def __subclasscheck__(self, subclass):
        from .serializers import ModelSerializer

        return issubclass(subclass, ModelSerializer)


def autodiscover():
    autodiscover_modules("rest", register_to=None)

    if getattr(settings, "WQ_CONFIG_FILE", None):
        with open(settings.WQ_CONFIG_FILE, "w") as f:
            dump_config(f, format="esm")


def register(*models, router=router, **kwargs):
    def register_decorator(serializer):
        if hasattr(serializer, "Meta"):
            model = getattr(serializer.Meta, "model", None)
            if model and model not in models:
                models.append(model)
        for model in models:
            router.register(model, serializer=serializer, **kwargs)
        return serializer

    return register_decorator



from django.db import models
from django.utils.translation import gettext_lazy as _
from wq.db.rest.models import LabelModel
import time


class RootModel(LabelModel):
    slug = models.SlugField()
    description = models.TextField()

    wq_label_template = "{{slug}}"


class OneToOneModel(LabelModel):
    root = models.OneToOneField(RootModel, models.CASCADE)

    wq_label_template = "onetoonemodel for {{root}}"


class ForeignKeyModel(LabelModel):
    root = models.ForeignKey(RootModel, models.CASCADE)

    wq_label_template = "foreignkeymodel for {{root}}"


class ExtraModel(LabelModel):
    root = models.ForeignKey(
        RootModel,
        models.CASCADE,
        related_name="extramodels",
    )
    alt_root = models.ForeignKey(
        RootModel,
        models.CASCADE,
        related_name="extramodel_set",
        null=True,
        blank=True,
    )

    wq_label_template = "extramodel for {{root}}"


class UserManagedModel(models.Model):
    user = models.ForeignKey("auth.User", models.CASCADE)


class Parent(LabelModel):
    name = models.CharField(max_length=10)


class Child(LabelModel):
    name = models.CharField(max_length=10)
    parent = models.ForeignKey(Parent, models.CASCADE, related_name="children")


class ItemType(LabelModel):
    name = models.CharField(max_length=10)
    active = models.NullBooleanField(default=True)


class Item(LabelModel):
    name = models.CharField(max_length=10)
    type = models.ForeignKey(ItemType, models.CASCADE)

    class Meta:
        ordering = ("type", "name")


class FileModel(LabelModel):
    name = models.CharField(max_length=255)
    file = models.FileField(upload_to="files")


class ImageModel(LabelModel):
    name = models.CharField(max_length=255)
    image = models.ImageField(upload_to="files")


class SlugModel(LabelModel):
    code = models.SlugField()
    name = models.CharField(max_length=255)


class SlugRefParent(LabelModel):
    ref = models.ForeignKey(SlugModel, models.CASCADE)
    name = models.CharField(max_length=255)

    wq_label_template = "{{name}}{{#ref_id}} ({{ref}}){{/ref_id}}"


class SlugRefChild(models.Model):
    parent = models.ForeignKey(SlugRefParent, models.CASCADE)
    name = models.CharField(max_length=255)


class DateModel(models.Model):
    name = models.CharField(max_length=10)
    date = models.DateTimeField()
    empty_date = models.DateTimeField(null=True)

    def __str__(self):
        return "%s on %s" % (self.name, self.date.date())


class ChoiceModel(LabelModel):
    CHOICE_CHOICES = [
        ("one", "Choice One"),
        ("two", "Choice Two"),
        ("three", "Choice Three"),
    ]
    GROUPED_CHOICES = [
        (
            "Group 1",
            (
                ("one", "Choice One"),
                ("two", "Choice Two"),
                ("three", "Choice Three"),
            ),
        )
    ]
    name = models.CharField(
        max_length=10,
        help_text="Enter Name",
    )
    choice = models.CharField(
        max_length=10,
        help_text="Pick One",
        choices=CHOICE_CHOICES,
    )
    grouped_choice = models.CharField(
        max_length=10,
        help_text="Pick One",
        choices=GROUPED_CHOICES,
    )

    wq_label_template = "{{name}}: {{choice}}"


class TranslatedModel(models.Model):
    name = models.CharField(
        verbose_name=_("translated model name"),
        max_length=255,
    )


class CharFieldModel(models.Model):
    required_field = models.CharField(max_length=20)
    nullable_field = models.CharField(null=True, max_length=20)
    blankable_field = models.CharField(blank=True, max_length=20)
    nullableblankable_field = models.CharField(
        null=True, blank=True, max_length=20
    )


class ExpensiveField(models.CharField):
    def __init__(self, cost=None, **kwargs):
        self.cost = cost
        super().__init__(**kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs["cost"] = self.cost
        return name, path, args, kwargs

    def to_python(self, value):
        if value:
            time.sleep(self.cost)
        return super().to_python(value)


class ExpensiveModel(LabelModel):
    name = models.CharField(max_length=20)
    expensive = ExpensiveField(cost=2, max_length=255)
    more_expensive = ExpensiveField(
        cost=5, max_length=255, null=True, blank=True
    )


class FieldsetModel(LabelModel):
    # General
    name = models.CharField(max_length=50)
    title = models.CharField(max_length=20)

    # Address
    address = models.CharField(max_length=255)
    city = models.CharField(max_length=255)



from django.db import models
from wq.db.rest.models import LabelModel


class Campaign(LabelModel):
    slug = models.SlugField()
    name = models.CharField(max_length=10)


class Attribute(LabelModel):
    name = models.CharField(max_length=10)
    campaign = models.ForeignKey(
        Campaign, models.CASCADE, related_name="attributes"
    )
    is_active = models.BooleanField(default=True)


class Entity(LabelModel):
    name = models.CharField(max_length=10)
    campaign = models.ForeignKey(Campaign, models.CASCADE)

    class Meta:
        verbose_name_plural = "entities"


class Value(models.Model):
    attribute = models.ForeignKey(Attribute, models.CASCADE)
    entity = models.ForeignKey(Entity, models.CASCADE, related_name="values")
    value = models.TextField()
